import csv
from pathlib import Path

g_list = Path("../species.txt")
ft_dir = Path("../feature")
gt_dir = Path("../gene")
ens_fields = ("Protein ID", "Gene ID", "Region Type", "Region",
              "Start", "End", "Strand", "Label", "Description")


def main():
    with open(g_list, "rt", newline="") as infile:
        reader = csv.DictReader(infile, delimiter="\t")
        sp2ft = {genome["filename"]: genome["feature"] for genome in reader}
    for species, ft_file in sp2ft.items():
        print(species, end="")
        outfile_path = gt_dir / (species + ".tsv")
        if outfile_path.exists():
            print(" - exists")
            continue
        print()
        gbk2ens_file(ft_dir / ft_file, outfile_path)


def gbk2ens_file(infile_name, outfile_name):
    with open(infile_name, "rt", newline="") as infile, \
         open(outfile_name, "w", newline="") as outfile:
        assert infile.read(2) == "# "
        reader = csv.DictReader(infile, delimiter="\t")
        writer = csv.DictWriter(outfile, delimiter="\t", fieldnames=ens_fields)
        outfile.write("\t".join(ens_fields) + "\n")
        genes = {}
        for feature in reader:
            if feature["assembly_unit"] != "Primary Assembly":
                if "GCF_000001635" not in str(infile_name):  # weird mouse :(
                    continue
            if feature["feature"] == "gene" and feature["class"] == "protein_coding":
                genes[feature["GeneID"]] = feature
            if feature["feature"] == "CDS" and feature["class"] == "with_protein":
                writer.writerow(gbk2ens(feature, genes[feature["GeneID"]]))


def gbk2ens(feature, gene):
    assert feature["GeneID"] == gene["GeneID"]
    if int(feature["start"]) < int(gene["start"]) or int(feature["end"]) > int(gene["end"]):
        print(f" (!) CDS outside of gene: {feature['product_accession']}")
    region = (
        feature["chromosome"]
        if feature["seq_type"] == "chromosome"
        else feature["genomic_accession"]
    )
    return {
        "Protein ID": feature["product_accession"],  # .split('.')[0],
        # 'Transcript ID': feature['related_accession'].split('.')[0],
        "Gene ID": feature["GeneID"],
        "Region Type": feature["seq_type"],
        "Region": region,
        # "Start": feature["start"],
        # "End": feature["end"],
        "Start": gene["start"],
        "End": gene["end"],
        "Strand": {"+": "1", "-": "-1"}[feature["strand"]],
        "Label": feature["symbol"],
        # "Description": feature["name"],
        "Description": gene["name"] or feature["name"],
    }


if __name__ == "__main__":
    main()
