// addspecies : data read functions
#include "addspecies.h"

// Split string at asterisk into path and extention
void    ParseExtension(const string& str, string& path, string &ext) 
{
    string::size_type pos = str.find_last_of('*');
    if (pos == string::npos) 
    {
        path.clear();
        ext = str;
    }
    else 
    {
        ext = str.substr(++pos);
        pos = str.find_first_of('*');
        path = str.substr(0, pos);
    }
}

// Split string at two asterisks into path, separator, and extension
void    ParseExtension(const string& str, string& path, string &ext, string &separ) 
{
    string::size_type pos = str.find_last_of('*');
    if (pos == string::npos) 
    {
        path.clear();
        separ.clear();
        ext = str;
    }
    else 
    {
        string::size_type pos1 = str.find_first_of('*');
        path = str.substr(0, pos1);
        pos1++;
        separ = pos1>=pos ? "" : str.substr(pos1, pos-pos1);
        ext = str.substr(pos+1);
    }
}

// Returns true at empty or comment line
bool    IsComment(const char* c) 
{
    while (*c != 0 && *c == ' ')
        c++;
    return (*c == 0 || *c == ';' || *c == '/' && *(c+1) == '/');
}

// Returns true at empty or comment line
bool    IsComment(string & s) 
{
    string::size_type pos = s.find_first_not_of(" \t");
    return (pos != string::npos && (s[pos]==';' || s.substr(pos,2)=="//"));
}

// Check if given pair of genes already occured.
bool    IsNewPair(const PGene pg1, const PGene pg2) 
{
    static PGene old1 = 0; 
    static PGene old2 = 0;
    if (pg1==old1 && pg2==old2)
        return false;
    else 
    {
        old1=pg1; old2=pg2;
    }
    vector <Homolog>::iterator pgit;
    vector <Homolog> & vpg1 = pg1->homologs[pg2->sp];
    for (pgit = vpg1.begin(); pgit != vpg1.end(); pgit++)
        if (pgit->pgene == pg2) return false;
    vector <Homolog> & vpg2 = pg2->homologs[pg1->sp];
    for (pgit = vpg2.begin(); pgit != vpg2.end(); pgit++)
        if (pgit->pgene == pg1) return false;
    return true;
}

// Return current token of the string str starting from offset.
// Modify the offset so that it points to the next token.
// Skip multiple tabs unless strict=true
// In non-strict mode, provide for a comment instead of the next token.
string  GetToken(const string & str, string::size_type & offset, bool strict = false) 
{
    static string token;
    string::size_type pos, pos2;
    while (true) 
    {
        pos = str.find_first_of('\t', offset);
        if (pos == string::npos)            // no more tabs
        {
            if (offset < str.size()) 
            {
                pos2 = str.find_first_not_of(' ', offset);
                if (pos2 != string::npos)
                    token = str.substr(pos2);
                else
                    token.clear();
            }
            else
                token.clear();
            offset = str.size();
            break;
        }
        else                                // found tab at pos position
        {
            pos2 = str.find_first_not_of(' ', offset);

            if (strict && pos2 == pos)          // empty or blank token
            {
                offset = pos+1;
                token.clear();
                return token;
            }
            else if (pos2 < pos)                // nonempty token
            {
                offset = pos2;
                token = str.substr(offset, pos-offset);
                offset = pos+1;
                break;
            }
            else offset = pos+1;                // skip repeating tabs
        }
    }
    if (!strict && token.length()>1 && token.substr(0,2)=="//") 
        token.clear();      // ignore comments
    return token;
}

// Read configuration file
bool    ReadConfig(void) 
{
    int lineno = 0;
    char buf[buflen];
    ifstream cf;
    cf.open(global.config.c_str());
    if (!cf || cf.bad()) 
    {
        cerr << "Can't open " << global.config << endl;
        return false;
    }

    while (!cf.eof())       // prior to 1st section
    {
        cf.getline(buf, buflen);
        lineno++;
        if (cf.bad()) return false;
        if (cf.eof()) return false;
        if (IsComment(buf)) continue;
        if (*buf == '[') break;
        cerr << "Invalid configuration line " << lineno << endl;
        return false;
    }

    while (!cf.eof())       // start of section found
    {
        string section(buf+1);
        string::size_type pos = section.find_last_of(']');
        if (pos == string::npos) 
        {
            cerr << "Invalid section '" << section << "' in line " << lineno << endl;
            return false;
        }
        section.erase(pos);

        if (section == "data") 
        {
            bool mdefined = false;
            while (!cf.eof())           // [folders]
            {
                cf.getline(buf, buflen);
                lineno++;
                if (cf.bad()) return false;
                if (cf.eof()) break;
                if (IsComment(buf)) continue;
                if (*buf == '[') break;

                string str(buf);
                string::size_type off = 0;
                string intype = GetToken(str, off);
                string inname = GetToken(str, off);

                switch (intype[0]) 
                {
                    case 'I': case 'i':
                        ParseExtension(inname, global.info_path, global.info_ext);
                        break;
                    case 'O': case 'o':
                    case 'H': case 'h':
                        ParseExtension(inname, global.homology_path, global.homology_ext);
                        break;
                    case 'P': case 'p':
                        ParseExtension(inname, global.paralogs_path, global.paralogs_ext);
                        break;
                    case 'A': case 'a':
                        ParseExtension(inname, global.matrix_path, global.matrix_ext, global.matrix_separ);
                        break;
                    case 'M': case 'm':
                        ParseExtension(inname, global.outmatrix_path, global.outmatrix_ext, global.outmatrix_separ);
                        mdefined = true;
                        break;
                }
            }
            if (global.outmatrix && !mdefined) 
            {
                cerr << "Need to define M folder for matrix output." << endl;
                return false;
            }
        }
        else if (section == "fields") 
        {
            while (!cf.eof())           // [fields]
            {
                cf.getline(buf, buflen);
                lineno++;
                if (cf.bad()) return false;
                if (cf.eof()) break;
                if (IsComment(buf)) continue;
                if (*buf == '[') break;

                string str(buf);
                string::size_type off = 0;
                string f = GetToken(str, off);
                if ((int)global.myfields.size() < ncol)
                    global.myfields.push_back(f);
            }
        }
        else if (section == "species") 
        {
            while (!cf.eof())           // [species]
            {
                cf.getline(buf, buflen);
                lineno++;
                if (cf.bad()) return false;
                if (cf.eof()) break;
                if (IsComment(buf)) continue;
                if (*buf == '[') break;

                string str(buf);
                string::size_type off = 0;
                string name = GetToken(str, off);
                if (name[0] == '*') continue;
                Species sp(name);
                for (int i = 0; i < ncol; i++)
                    sp.myfields.push_back(-1);
                global.Animals.push_back(sp);
                global.nAnimals++;
            }
        }
        else if (section == "add")
        {
            while (!cf.eof())           // [add]
            {
                cf.getline(buf, buflen);
                lineno++;
                if (cf.bad()) return false;
                if (cf.eof()) break;
                if (IsComment(buf)) continue;
                if (*buf == '[') break;

                string str(buf);
                string::size_type off = 0;
                string key = GetToken(str, off);
                string val = GetToken(str, off);
                if (!global.givenElimit && key.compare(0,4,"EVAL")==0)
                {
                    istringstream iss(val);
                    iss >> global.Elimit;
                }
                else if (!global.givenOweight && key.compare(0,4,"ORTH")==0)
                {
                    istringstream iss(val);
                    iss >> global.oweight;
                }
                else if (!global.givenPweight && key.compare(0,4,"PARA")==0)
                {
                    istringstream iss(val);
                    iss >> global.pweight;
                }
            }
        }
        else
        {
            while (!cf.eof())           // [unknown]
            {
                cf.getline(buf, buflen);
                lineno++;
                if (cf.bad()) return false;
                if (cf.eof()) break;
                if (IsComment(buf)) continue;
                if (*buf == '[') break;
            }
        }
    } 
    cf.clear(); cf.close();
    return true;
}

// Read table of genes for given species no.
bool    ReadGeneInfo(int sp) 
{
    string filename = global.info_path + global.Animals[sp].name + global.info_ext;
    char buf[buflen];
    ifstream df;
    df.open(filename.c_str());
    if (!df || df.bad()) 
    {
        cout << "Can't open " << filename << endl;
        return false;
    }

    df.getline(buf, buflen);      // Read and check heading line
    string str(buf);
    string::size_type offset = 0;

    int field_no;
    string name;

    for (field_no = 0; ; field_no++)       // this species columns
    {
        name = GetToken(str, offset, true);
        if (name.empty()) break;

        int j;
        for (j = 0; j < (int)global.myfields.size(); j++) 
        {
            if (name == global.myfields[j]) 
            {
                global.Animals[sp].myfields[j] = field_no;
                global.Animals[sp].required.push_back(j);
                break;
            }
        }
        if (j >= (int)global.myfields.size())
            global.Animals[sp].required.push_back(-1);

        global.Animals[sp].fields.push_back(name);
    }
    // Check that all required sp fields present
    for (int j = 0; j < ncol; j++) 
    {
        if (global.Animals[sp].myfields[j] < 0) 
        {
            cout << "Required field '" << global.myfields[j] << "' of " << global.Animals[sp].name
                << " is absent in " << filename << endl;
            return false;
        }
    }

    // Read and parse protein/gene lines
    string protein;
    Species& species = global.Animals[sp];
    Giter    git;
    Gene     gene;
    if (sp == global.tosp)  // add gene.homologs[nAnimals] for self-scores
    {
        vector <Homolog> selfhomo;
        gene.homologs.push_back(selfhomo);
    }
    int line=1, nprot=0, ngenes=0;

    while (!df.eof()) 
    {
        df.getline(buf, buflen);
        line++;
        if (df.bad()) return false;             // IO error
        if (df.eof() && *buf == 0) break;       // EOF or empty line

        // Read current line & store gene information
        str.assign(buf);
        offset = 0;

        for (field_no = 0; field_no < (int)species.fields.size(); field_no++) 
        {                             
            name = GetToken(str, offset, true);
            switch (species.required[field_no]) 
            {
                case field_protein:
                    protein = name;         // Protein
                    break;
                case field_gene:
                    gene.name = name;       // Gene
                    break;
            }
        }

        gene.sp = sp;                  // Species - back reference

        if (protein.empty()) continue;      // skip genes w/o protein
        nprot++;

        if (!gene.name.empty()) 
        {
            // add gene if needed and find pgene
            git = species.genes.find(gene.name);
            if (git == species.genes.end()) 
            {
                ngenes++;
                git = species.genes.insert( git, Gpair(gene.name, gene) );
            }
            PGene pgene = (Gene*)&git->second;

            // map gene id -> pgene
            Piter piter = global.stableid.find(gene.name);
            if (piter == global.stableid.end())
                global.stableid.insert( Ppair(gene.name, pgene) );

            // map protein id -> pgene
            piter = global.stableid.find(protein);
            if (piter == global.stableid.end())
                global.stableid.insert( Ppair(protein, pgene) );
            else
                cout << "Duplicate protein " << protein << " in line " << line << endl;
        }

        if (inrecordstep && line % inrecordstep == 0) 
        {
            cout << line << "    \r";
        }
    }

    cout << "Lines read: " << line << "  genes: " << ngenes << "  proteins: " << nprot << endl;
    df.clear(); df.close();
    return true;
}

// Read table of homologs for genes of a given species.
// Use table of paralogs if paralogs=true.
bool    ReadShortData(int sp, bool paralogs) 
{
    int npairs = 0;
    int sp1 = sp;
    char buf[buflen];
    string filename = paralogs ? global.paralogs_path + global.Animals[sp1].name + global.paralogs_ext
                               : global.homology_path + global.Animals[sp1].name + global.homology_ext;
    string str(buf);
    string::size_type offset = 0;
    ifstream df;
    df.open(filename.c_str());
    if (!df || df.bad()) 
    {
        cout << "Can't open " << filename << endl;
        return false;
    }

    Species& species1 = global.Animals[sp1];

    for (int line = 1; !df.eof(); line++) 
    {
        df.getline(buf, buflen);
        if (df.bad())
            return false;               // IO error
        if (df.eof() && *buf == 0)
            break;                      // EOF or empty line

        // Read & parse current pair
        str.assign(buf);
        offset = 0;
        string name1 = GetToken(str, offset, true);
        string name2 = GetToken(str, offset, true);
        if (name1.empty() || name2.empty())
            continue;                   // skip improper lines

        Piter piter = global.stableid.find(name1);
        if (piter == global.stableid.end())
            continue;                   // skip lines with unknown name1
        PGene pg1 = piter->second;

        piter = global.stableid.find(name2);
        if (piter == global.stableid.end())
            continue;                   // skip lines with unknown name2
        PGene pg2 = piter->second;

        // Store pairing in both genes (only if new)
        if (IsNewPair(pg1, pg2)) 
        {
            pg1->homologs[pg2->sp].push_back( Homolog(pg2,0,maxint) );
            pg2->homologs[pg1->sp].push_back( Homolog(pg1,0,maxint) );
            npairs++;
        }

        if (inmatrixstep && line % inmatrixstep == 0)
            cout << back9 << setw(9) << line;
    }
    cout << back9 << setw(9) << npairs;
    df.clear(); df.close();
    return true;
}

// Read matrix of scores between two species
bool    ReadMatrix(int sp1, int sp2) 
{
    char buf[buflen];
    int line; 
    int npairs = 0;
    string::size_type offset;
    string filename = global.matrix_path + global.Animals[sp1].name + 
        global.matrix_separ + global.Animals[sp2].name + global.matrix_ext;

    ifstream df;
    df.open(filename.c_str());
    if (!df || df.bad()) 
    {
        cout << "Can't open " << filename << endl;
        return false;
    }

    string lastprot1;
    Piter pit1;
    for (line=1; !df.eof(); line++) 
    {
        df.getline(buf, buflen);
        if (df.bad()) return false;             // IO error
        if (df.eof() && *buf == 0) break;       // EOF or empty line

        // Read & parse current pair
        string str(buf);
        string prot1, prot2;
        offset = 0;

        prot1 = GetToken(str, offset, true);    // Protein 1
        if (prot1.empty()) 
        {
            cout << "Field 1 error in line " << line << endl;
            continue;
        }
        if (prot1 != lastprot1) 
        {
            pit1 = global.stableid.find(prot1);
            if (pit1 == global.stableid.end())
                continue;
            lastprot1 = prot1;
        }
        PGene pg1 = pit1->second;
        int sp = pg1->sp;
        if (sp1 != sp) 
        {
            cout << "Species1=" << sp << " encountered in line " << line << endl;
            continue;
        }

        prot2 = GetToken(str, offset, true);    // Protein 2
        if (prot2.empty()) 
        {
            cout << "Field 2 error in line " << line << endl;
            continue;
        }
        Piter pit2 = global.stableid.find(prot2);
        if (pit2 == global.stableid.end())
            continue;           // skip unknown or alien proteins
        PGene pg2 = pit2->second;
        sp = pg2->sp;
        if (sp2 != sp) 
        {
            cout << "Species2=" << sp << " encountered in line " << line << endl;
            continue;
        }

        string rscore = GetToken(str, offset, true);  // Raw score
        int rawscore = -1;
        istringstream rs(rscore);
        rs >> rawscore;
        if (rawscore < 0 || rscore.empty()) 
        {
            cout << "Score invalid or absent from line " << line << endl;
            continue;
        }

        string evalue = GetToken(str, offset, true);  // E-value
        double expect = -1;
        istringstream es(evalue);
        es >> expect;
        if (expect < 0 || evalue.empty()) 
        {
            cout << "E-value invalid or absent from line " << line << endl;
            continue;
        }
        if (pg1 != pg2 && expect > global.Elimit ||
            pg1 == pg2 && expect > global.Elimit * lower_E)
                continue;       // skip homologs with too great E-value

        int sq2 = (sp1==sp2 && sp2==global.tosp) ? global.nAnimals : sp2;
        vector <Homolog> & vhg1 = pg1->homologs[sq2];    // sp2 | nAnimals
        vector <Homolog>::iterator hgit1 = vhg1.begin();
        for ( ; hgit1 != vhg1.end(); hgit1++) 
        {
            if (hgit1->pgene == pg2)             // modify old pair if needed
            {
                if (hgit1->evalue > expect)  hgit1->evalue = expect;
                if (hgit1->score < rawscore) hgit1->score  = rawscore;
                break;
            }
        }

        int sq1 = (sp1==sp2 && sp1==global.tosp) ? global.nAnimals : sp1;
        vector <Homolog> & vhg2 = pg2->homologs[sq1];    // sp1 | nAnimals
        vector <Homolog>::iterator hgit2 = vhg2.begin();
        for ( ; hgit2 != vhg2.end(); hgit2++) 
        {
            if (hgit2->pgene == pg1)             // modify old pair if needed
            {
                if (hgit2->evalue > expect)  hgit2->evalue = expect;
                if (hgit2->score < rawscore) hgit2->score  = rawscore;
                break;
            }
        }

        if (hgit1 == vhg1.end() && hgit2 == vhg2.end())    // add new pair(s) if needed
        {
            pg1->homologs[sq2].push_back( Homolog(pg2, expect, rawscore) );
            if (pg2 != pg1)
                pg2->homologs[sq1].push_back( Homolog(pg1, expect, rawscore) );
            npairs ++;
        }
        else if (hgit1 == vhg1.end()) 
        {
            pg1->homologs[sq2].push_back( Homolog(pg2, expect, rawscore) );
        }
        else if (hgit2 == vhg2.end()) 
        {
            pg2->homologs[sq1].push_back( Homolog(pg1, expect, rawscore) );
        }

        if (inmatrixstep && line % inmatrixstep == 0)
            cout << back9 << setw(9) << line;
    }

    cout << back9 << setw(9) << npairs;
    df.clear(); df.close();
    return true;
}

// Return whether 1st homolog is better than 2nd
bool    GreaterScore(const Homolog& h1, const Homolog& h2) 
{
    return h1.score > h2.score;
}

// Return whether 1st gene precedes 2nd one
bool    LessID(const PGene p1, const PGene p2)
{
    return p1->name < p2->name;
}

// Write matrix of intergenic scores between two species
bool    WriteMatrix(int sp1, int sp2) 
{
    int npairs = 0;
    //int sq1 = (sp1==sp2 && sp1==global.tosp) ? global.nAnimals : sp1;
    int sq2 = (sp1==sp2 && sp2==global.tosp) ? global.nAnimals : sp2;
    string filename = global.outmatrix_path + global.Animals[sp1].name + 
        global.outmatrix_separ + global.Animals[sp2].name + global.outmatrix_ext;
    ofstream df;
    df.open(filename.c_str());
    if (!df || df.bad()) 
    {
        cout << "Can't create " << filename << endl;
        return false;
    }

    Species & spec = global.Animals[sp1];
    Giter git = spec.genes.begin();
    for ( ; git != spec.genes.end(); git++) 
    {
        Gene & gene = git->second;
        vector <Homolog> & vh = gene.homologs[sq2];     // instead of sp2
        if (vh.empty()) continue;   // skip genes w/o homologs
        if (!global.unsorted)
            sort(vh.begin(), vh.end(), GreaterScore);
        vector <Homolog>::iterator hit = vh.begin();
        for ( ; hit != vh.end(); hit++) 
        {
            df << gene.name << "\t" << hit->pgene->name << "\t" << hit->score << "\t";
            if (hit->evalue == 0.0) 
                df << fixed << setprecision(1) << hit->evalue << endl; 
            else if (hit->evalue >= 0.001)
                df << fixed << setprecision(3) << hit->evalue << endl; 
            else 
                df << scientific << setprecision(3) << hit->evalue << fixed << endl;
            if (outmatrixstep != 0 && npairs % outmatrixstep == 0)
                cout << back9 << setw(9) << npairs;
            npairs++;
        }
    }

    cout << back9 << setw(9) << npairs;
    df.clear(); df.close();
    return true;
}
